"""
Eval functions that be used/reused with Prompts

All functions here are expected to:
* receive a response string
* [optional] receive arbitrary extra context
* return a value from 0 to 1, 0 representing failing at the task, 1 full
success, and a range in-between
"""

from typing import List


def percentage_of_words(response: str, words: List[str], case_sensitive: bool = False) -> float:
    """
    Calculate the percentage of words from a list that are present in the given response.

    Args:
        response (str): The string to search for words.
        words (List[str]): A list of words to check for their presence in the response.
        case_sensitive (bool, optional): If True, the search will be case-sensitive;
                                         otherwise, it will be case-insensitive. Defaults to False.

    Returns:
        float: The percentage of words found in the response (0.0 to 1.0).

    Examples:
    >>> percentage_of_words_in_response("This is an Example string.", ["example", "test"])
    0.5
    >>> percentage_of_words_in_response(
        "This is an Example string.", ["Example"], case_sensitive=True)
    1.0
    >>> percentage_of_words_in_response(
        "This is an Example string.", ["example"], case_sensitive=True)
    0.0
    >>> percentage_of_words_in_response("This is an Example string.", ["notfound"])
    0.0
    """
    if not case_sensitive:
        response = response.lower()
        words = [w.lower() for w in words]

    return sum(1 if w in response else 0 for w in words) / len(words)


def _common_word_search(
    response: str, words: List[str], case_sensitive: bool, match_type: str
) -> int:
    if not case_sensitive:
        response = response.lower()
        words = [w.lower() for w in words]

    if match_type == "any":
        return 1 if any(w in response for w in words) else 0
    elif match_type == "all":
        return 1 if all(w in response for w in words) else 0
    else:
        raise ValueError("Invalid match_type. Accepted values are 'any' or 'all'.")


def any_word(response: str, words: List[str], case_sensitive: bool = False) -> int:
    """
    Check if any word from a list of words is present in the given response and return 1 or 0.

    Args:
        response (str): The string to search for words.
        words (List[str]): A list of words to check for their presence in the response.
        case_sensitive (bool, optional): If True, the search will be case-sensitive;
                                         otherwise, it will be case-insensitive. Defaults to False.

    Returns:
        int: 1 if any word from the list is found in the response; otherwise, 0.

    Examples:
    >>> any_word("This is an Example string.", ["example", "test"])
    1
    >>> any_word("This is an Example string.", ["Example"], case_sensitive=True)
    1
    >>> any_word("This is an Example string.", ["example"], case_sensitive=True)
    0
    >>> any_word("This is an Example string.", ["notfound"])
    0
    """
    return _common_word_search(response, words, case_sensitive, match_type="any")


def all_words(response: str, words: List[str], case_sensitive: bool = False) -> int:
    """
    Check if all words from a list of words are present in the given response and return 1 or 0.

    Args:
        response (str): The string to search for words.
        words (List[str]): A list of words to check for their presence in the response.
        case_sensitive (bool, optional): If True, the search will be case-sensitive;
                                         otherwise, it will be case-insensitive. Defaults to False.

    Returns:
        int: 1 if all words from the list are found in the response; otherwise, 0.

    Examples:
    >>> all_words("This is an Example string with test.", ["example", "test"])
    1
    >>> all_words("This is an Example string with Test.", ["Example", "Test"], case_sensitive=True)
    1
    >>> all_words("This is an Example string.", ["example", "test"], case_sensitive=True)
    0
    >>> all_words("This is an Example string.", ["example", "notfound"])
    0
    """
    return _common_word_search(response, words, case_sensitive, match_type="all")



"""
Some basic examples for promptimize.

to run, simply execute `p9e ./examples/readme_examples.py`
"""
# Brining some "prompt generator" classes
from promptimize.prompt_cases import PromptCase

# Bringing some useful eval function that help evaluating and scoring responses
# eval functions have a handle on the prompt object and are expected
# to return a score between 0 and 1
from promptimize import evals

# Promptimize will scan the target folder and find all Prompt objects
# and derivatives that are in the python modules
simple_prompts = [
    # Prompting "hello there" and making sure there's "hi" or "hello"
    # somewhere in the answer
    PromptCase("hello there!", lambda x: evals.any_word(x.response, ["hi", "hello"])),
    PromptCase(
        "name the top 10 guitar players!",
        lambda x: evals.all_words(x.response, ["zappa", "hendrix"]),
        weight=2,
    ),
]



import jinja2

environment = jinja2.Environment()


def process_template(template_as_string, **kwargs):
    template = environment.from_string(template_as_string)
    return template.render(**kwargs)



import os
from typing import Any, Callable, List, Optional, Union

from langchain.llms import OpenAI

from box import Box

from promptimize import utils
from promptimize.simple_jinja import process_template


class BasePromptCase:
    """Abstract base prompt case"""

    attributes_used_for_hash = set()
    verbose_attrs = {"prompt"}

    def __init__(
        self,
        evaluators: Optional[Union[Callable, List[Callable]]] = None,
        key: Optional[str] = None,
        weight=1,
        category: str = None,  # used for info/reporting purposes only
        prompt_executor: Any = None,
        prompt_executor_kwargs: dict = None,
        prompt_hash=None,
        *args,
        **kwargs,
    ) -> None:
        """
        Initialize a SimplePromptCase instance.

        Args:
            user_input (str): Raw user_input for the prompt.
            evaluators (Optional[Union[Callable, List[Callable]]]): Optional
                callable or list of callables used for evaluation.
            key (Optional[str]): Optional unique key for the prompt.
            weight (int, optional): Optional weight for the prompt (default: 1).
            category (Optional[str], optional): Optional category for
                the prompt (used for info/reporting purposes only).
        """
        self.extra_args = args
        self.extra_kwargs = kwargs
        self.response = None
        self.has_run = False
        self.was_tested = False
        self.test_results = None
        self.evaluators = evaluators or []
        self.weight = weight or 1
        self.category = category
        self.pre_run_output = None
        self.post_run_output = None
        self.prompt_executor = prompt_executor or self.get_prompt_executor()
        self.prompt_executor_kwargs = prompt_executor_kwargs or {}

        self._prompt_hash = prompt_hash

        self.execution = Box()

        self.prompt = utils.literal_str(self.render()).strip()

        self.key = key or "prompt-" + self.prompt_hash

        if not utils.is_iterable(self.evaluators):
            self.evaluators = [self.evaluators]  # type: ignore

    def get_prompt_executor(self):
        model_name = os.environ.get("OPENAI_MODEL") or "text-davinci-003"
        openai_api_key = os.environ.get("OPENAI_API_KEY")
        self.prompt_executor_kwargs = {"model_name": model_name}
        return OpenAI(model_name=model_name, openai_api_key=openai_api_key)

    def execute_prompt(self, prompt_str):
        self.response = self.prompt_executor(prompt_str)
        return self.response

    def pre_run(self):
        pass

    def post_run(self):
        pass

    @property
    def hash(self):
        return utils.short_hash(hash(self))

    def __hash__(self):
        attrs = self.attributes_used_for_hash
        s = "|".join([utils.short_hash(utils.hashable_repr(getattr(self, attr))) for attr in attrs])
        return utils.int_hash(s)

    def render(self):
        raise NotImplementedError()

    def get_unique_hash(self, extra_context=None):
        """Returns a unique identifier, determined by the run

        Generally, the actual call sent to GPT (prompt, execution params)
        represent something unique.
        """
        return utils.short_hash(str(self.extra_kwargs))

    def to_dict(self, verbose=False):
        d = {
            "key": self.key,
            "prompt_hash": self.prompt_hash,
            "prompt": self.prompt,
            "response": self.response,
            "weight": self.weight,
            "execution": self.execution.to_dict(),
        }
        return d

    def print(self, verbose=False, style="yaml"):
        style = style or "yaml"
        output = self.to_dict(verbose)
        if not verbose:
            for attr in self.verbose_attrs:
                del output[attr]
        if self.weight == 1:
            del output["weight"]
        highlighted = utils.serialize_object(output, style)
        print(highlighted)

    def test(self):
        test_results = []
        for evaluator in self.evaluators:
            result = evaluator(self)
            if not (utils.is_numeric(result) and 0 <= result <= 1):
                raise Exception("Value should be between 0 and 1")
            test_results.append(result)

        if len(test_results):
            self.execution.score = sum(test_results) / len(test_results)
        self.was_tested = True

    @property
    def prompt_hash(self):
        if self._prompt_hash:
            return self._prompt_hash
        return utils.short_hash(hash(self))

    def _run(self, dry_run):
        pre_run_output = self.pre_run()
        if pre_run_output:
            self.execution.pre_run_output = pre_run_output

        if not dry_run:
            with utils.MeasureDuration() as md:
                self.response = self.execute_prompt(self.prompt).strip()

            self.execution.api_call_duration_ms = md.duration

            post_run_output = self.post_run()
            if post_run_output:
                self.execution.post_run_output = post_run_output
            self.has_run = True
            self.execution.run_at = utils.current_iso_timestamp()
            return self.response


class PromptCase(BasePromptCase):
    """A simple prompt case"""

    attributes_used_for_hash = BasePromptCase.attributes_used_for_hash | {"user_input"}

    def __init__(
        self,
        user_input,
        *args,
        **kwargs,
    ) -> None:
        self.user_input = user_input
        super().__init__(*args, **kwargs)

    def to_dict(self, verbose=False, *args, **kwargs):
        d = super().to_dict(*args, **kwargs)
        d = utils.insert_in_dict(d, "user_input", self.user_input, after_key="key")
        return d

    def render(self):
        return self.user_input


class TemplatedPromptCase(BasePromptCase):
    template_defaults: dict = {}
    attributes_used_for_hash = BasePromptCase.attributes_used_for_hash | {
        "user_input",
        "extra_kwargs",
    }
    template = "{{ user_input }}"

    def __init__(
        self,
        user_input=None,
        *args,
        **kwargs,
    ) -> None:
        self.user_input = user_input
        return super().__init__(*args, **kwargs)

    def to_dict(self, verbose=False, *args, **kwargs):
        d = super().to_dict(*args, **kwargs)
        d = utils.insert_in_dict(d, "user_input", self.user_input, after_key="key")
        return d

    def get_extra_template_context(self):
        """meant to be overriden in derived classes to add logic/context"""
        return {}

    @property
    def jinja_context(self):
        context_kwargs = self.template_defaults.copy()
        context_kwargs.update(self.get_extra_template_context())
        context_kwargs.update(self.extra_kwargs)
        context_kwargs.update({"user_input": self.user_input})
        return context_kwargs

    def render(self, **kwargs):
        return process_template(self.template, **self.jinja_context)


class LangchainPromptCase(BasePromptCase):
    attributes_used_for_hash = BasePromptCase.attributes_used_for_hash | {
        "extra_kwargs",
        "langchain_prompt",
    }

    def __init__(
        self,
        langchain_prompt,
        *args,
        **kwargs,
    ) -> None:
        self.langchain_prompt = langchain_prompt
        return super().__init__(*args, **kwargs)

    def to_dict(self, verbose=False, *args, **kwargs):
        d = super().to_dict(*args, **kwargs)
        d = utils.insert_in_dict(d, "prompt_kwargs", self.extra_kwargs, after_key="key")
        return d

    def render(self):
        return self.langchain_prompt.format(**self.extra_kwargs)



import json
from typing import Any, Dict, List, Optional
import subprocess
import hashlib
from datetime import datetime
import re
import time
from pygments import highlight
from pygments.lexers import YamlLexer, JsonLexer
from pygments.formatters import TerminalFormatter
from tabulate import tabulate
import yaml
from yaml.representer import SafeRepresenter


def is_numeric(value):
    """that'd be nice if we had this in the std lib..."""
    return isinstance(value, (int, float, complex))


def is_iterable(obj):
    """that'd be nice if we had this in the std lib..."""
    try:
        iter(obj)
        return True
    except TypeError:
        return False


# Yaml-related section, hacking out of some issues


class folded_str(str):
    pass


class literal_str(str):
    pass


def change_style(style, representer):
    def new_representer(dumper, data):
        scalar = representer(dumper, data)
        scalar.style = style
        return scalar

    return new_representer


# represent_str does handle some corner cases, so use that
# instead of calling represent_scalar directly
represent_folded_str = change_style(">", SafeRepresenter.represent_str)
represent_literal_str = change_style("|", SafeRepresenter.represent_str)


def str_presenter(dumper, data):
    """
    Some hack to get yaml output to use look good for multiline,
    which is common in this package

    from: https://shorturl.at/klvC1
    """
    if len(data.splitlines()) > 1 or "\n" in data or "\\n" in data:
        text_list = [line.rstrip() for line in data.splitlines()]
        fixed_data = "\n".join(text_list)
        return dumper.represent_scalar("tag:yaml.org,2002:str", fixed_data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(folded_str, represent_folded_str)
yaml.add_representer(literal_str, represent_literal_str)
yaml.add_representer(str, str_presenter)


def extract_json_objects(text: str, get_first: bool = True) -> List[Dict[str, Any]]:
    """
    Extract JSON objects from a given string by looking for matching curly brackets.

    Args:
        text (str): The input string containing JSON objects.

    Returns:
        List[Dict[str, Any]]: A list of JSON objects found in the input string.

    Example:

    >>> extract_json_objects('Some text: {"key1": "value1", "key2": "value2"} and more text.')
    [{'key1': 'value1', 'key2': 'value2'}]

    >>> extract_json_objects('No JSON objects in this text.')
    []

    >>> extract_json_objects('{"a": 1, "b": 2} and {"c": 3, "d": 4}')
    [{'a': 1, 'b': 2}, {'c': 3, 'd': 4}]
    """
    json_objects = []
    # man that regex is something... generated by GPT4...
    regex = r"(\{(?:[^{}]|(?:\{(?:[^{}]|(?:\{(?:[^{}]|(?:\{[^{}]*\}))*\}))*\}))*\})"
    for match in re.finditer(regex, text):
        try:
            json_object = json.loads(match.group())
            json_objects.append(json_object)
        except json.JSONDecodeError:
            pass
    return json_objects


def int_hash(text):
    # Create a SHA-256 hash of the input string
    hash_object = hashlib.sha256(str(text).encode())

    # Convert the hash to a hexadecimal string
    hex_hash = hash_object.hexdigest()
    return int(hex_hash, 16)


def short_hash(obj, length=8):
    hash_object = hashlib.sha256(str(obj).encode())
    hex_hash = hash_object.hexdigest()
    return hex_hash[:length]


def to_yaml(data, highlighted=True):
    data = yaml.dump(data, sort_keys=False)
    if highlighted:
        data = highlight(data, YamlLexer(), TerminalFormatter())
    return data


def to_json(data, highlighted=True):
    data = json.dumps(data, indent=2)
    highlighted = highlight(data, JsonLexer(), TerminalFormatter())
    return highlighted


def serialize_object(data, style="yaml", highlighted=True):
    if style == "yaml":
        return to_yaml(data, highlighted)
    return to_json(data, highlighted)


def transform_strings(obj, transformation):
    """
    Recursively iterates through nested iterables (lists and tuples) and dictionaries,
    applying a given transformation function to all strings found.

    :param obj: The input object, which can be a string, dictionary, list, or tuple.
                Other types will be returned unchanged.
    :param transformation: A function that takes a single string argument and returns
                           a transformed string.
    :return: A new object with the same structure as the input object, but with all
             strings transformed by the given transformation function.
    """
    if isinstance(obj, str):
        return transformation(obj)
    elif isinstance(obj, dict):
        return {key: transform_strings(value, transformation) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [transform_strings(item, transformation) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(transform_strings(item, transformation) for item in obj)
    else:
        return obj


def get_git_info(sha_length: int = 12):
    try:
        sha = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()
        if sha_length:
            sha = sha[:sha_length]
        branch = (
            subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
            .decode("utf-8")
            .strip()
        )
        dirty = subprocess.call(["git", "diff-index", "--quiet", "HEAD"]) != 0

        return {"sha": sha, "branch": branch, "dirty": dirty}
    except subprocess.CalledProcessError:
        return None


class MeasureDuration:
    def __init__(self):
        self.duration = None

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        end_time = time.time()
        self.duration = (end_time - self.start_time) * 1000


def insert_in_dict(
    dictionary: Dict[Any, Any],
    key: Any,
    value: Any,
    position: Optional[int] = None,
    before_key: Optional[Any] = None,
    after_key: Optional[Any] = None,
) -> Dict[Any, Any]:
    """
    Insert a key/value pair in a dictionary at a specific position, before a
        specified key, or after a specified key.

    Args:
        dictionary (Dict[Any, Any]): The original dictionary.
        key (Any): The key to be inserted.
        value (Any): The value associated with the key.
        position (Optional[int], optional): The position at which the key/value pair
            should be inserted. Defaults to None.
        before_key (Optional[Any], optional): The key before which the new
            key/value pair should be inserted. Defaults to None.
        after_key (Optional[Any], optional): The key after which the new
            key/value pair should be inserted. Defaults to None.

    Raises:
        ValueError: If more than one of 'position', 'before_key', or 'after_key' is specified.
        ValueError: If the specified position is out of range.
        KeyError: If 'before_key' or 'after_key' is not found in the dictionary.

    Returns:
        Dict[Any, Any]: A new dictionary with the inserted key/value pair.
    """
    if sum([bool(position is not None), bool(before_key), bool(after_key)]) > 1:
        raise ValueError("Only one of 'position', 'before_key', or 'after_key' can be specified")

    if position is not None and (position > len(dictionary) or position < 0):
        raise ValueError("Position is out of range")

    if before_key is not None and before_key not in dictionary:
        raise KeyError(f"'before_key': {before_key} not found in the dictionary")

    if after_key is not None and after_key not in dictionary:
        raise KeyError(f"'after_key': {after_key} not found in the dictionary")

    new_dict = {}
    inserted = False

    for index, (dict_key, dict_value) in enumerate(dictionary.items()):
        if position == index or dict_key == before_key:
            new_dict[key] = value
            inserted = True
        elif after_key == dict_key:
            new_dict[dict_key] = dict_value
            new_dict[key] = value
            inserted = True
            continue

        new_dict[dict_key] = dict_value

    if not inserted or position == len(dictionary):
        new_dict[key] = value

    return new_dict


def current_iso_timestamp():
    now = datetime.utcnow()
    return now.isoformat()


def hashable_repr(obj):
    if isinstance(obj, (list, tuple, set)):
        return "".join(hashable_repr(item) for item in obj)
    elif isinstance(obj, dict):
        return "".join(
            hashable_repr(key) + hashable_repr(value) for key, value in sorted(obj.items())
        )
    elif callable(obj):
        return str(obj.__code__.co_code)
    else:
        return str(obj)


def trabulate(df, showindex=True, headers="keys"):
    headers = headers if headers else []
    for column in df.columns:
        if df[column].dtype == "int64":
            df[column] = df[column].astype(str)
    return tabulate(df, headers=headers, showindex=showindex, tablefmt="psql", floatfmt=".2f")



"""
Some basic examples for promptimize.

to run, simply execute `p9e ./examples/readme_examples.py`
"""
# Brining some "prompt generator" classes
from promptimize.prompt_cases import PromptCase, TemplatedPromptCase

# Bringing some useful eval function that help evaluating and scoring responses
# eval functions have a handle on the prompt object and are expected
# to return a score between 0 and 1
from promptimize import evals

# Promptimize will scan the target folder and find all Prompt objects
# and derivatives that are in the python modules
simple_prompts = [
    # Prompting "hello there" and making sure there's "hi" or "hello"
    # somewhere in the answer
    PromptCase("hello there!", lambda x: evals.any_word(x.response, ["hi", "hello"])),
    # Making sure 3 specific guitar players are in the top 10
    # the score here is a percentage of the words found
    PromptCase(
        "who are the top 10 best guitar players of all time?",
        lambda x: evals.percentage_of_words(
            x.response, ["frank zappa", "david gilmore", "carlos santana"]
        ),
    ),
    # GPT-ing myself and making sure the response mentions Superset and Airflow
    PromptCase(
        "who is Maxime Beauchemin, (the data guy...)?",
        lambda x: evals.percentage_of_words(
            x.response, ["superset", "airflow"], case_sensitive=False
        ),
    ),
]

# deriving TemplatedPromptCase to generate prompts that ask GPT to generate SQL
# based on table schemas. The point here is you can derive the `Prompt`
# class to create more specialized Prompt generators
# For instance, the SqlPropt class defined bellow could be extended to fetch
# schema definitions dynamically, acutally run the SQL, and allow
# doing evals against the resultset.


class SqlPrompt(TemplatedPromptCase):
    # the TemplatedPromptCase expects a dict of defaults that can be overriden in the constructor
    template_defaults = {"dialect": "Postgres"}
    # The actual Jinja2 template
    template = """\
    given these SQL table schemas:
        CREATE TABLE world_population (
            country_name STRING,
            year DATE,
            population_total INT,
        );

    So, can you write a SQL query for {{ dialect }} that answers this user prompt:
    {{ input }}
    """


# Generating a few SQL prompts
sql_prompts = [
    SqlPrompt(
        # you can pass a unique key that can we used to reference a prompt
        key="sql-top-10-inc",
        # the user input that'll be added in place of {{ input }} in the template above
        user_input="give me the top 10 countries with the highest net increase of population over the past 25 years?",
        # the dialect template parameter, overriding the default set above
        dialect="BigQuery",
        # a simple validation function making sure the SQL starts with SELECT
        evaluators=lambda x: 1 if x.response.strip().startswith("SELECT") else 0,
    ),
]



import click

from promptimize.crawler import discover_objects
from promptimize.prompt_cases import BasePromptCase
from promptimize.reports import Report
from promptimize.suite import Suite


@click.group(help="💡¡promptimize!💡 CLI. `p9e` works too!")
def cli():
    pass


@click.command(help="run some prompts")
@click.argument(
    "path",
    required=True,
    type=click.Path(exists=True),
)
@click.option("--verbose", "-v", is_flag=True, help="Trigger more verbose output")
@click.option("--force", "-f", is_flag=True, help="Force run, do not skip")
@click.option(
    "--human",
    "-h",
    is_flag=True,
    help="Human review, allowing a human to review and force pass/fail each prompt case",
)
@click.option("--repair", "-r", is_flag=True, help="Only re-run previously failed")
@click.option("--dry-run", "-x", is_flag=True, help="DRY run, don't call the API")
@click.option("--shuffle", is_flag=True, help="Shuffle the prompts in a random order")
@click.option(
    "--style",
    "-s",
    type=click.Choice(["json", "yaml"], case_sensitive=False),
    default="yaml",
    help="json or yaml formatting",
)
@click.option(
    "--max-tokens",
    "-m",
    type=click.INT,
    default=1000,
    help="max_tokens passed to the model",
)
@click.option(
    "--limit",
    "-l",
    type=click.INT,
    default=0,
    help="limit how many prompt cases to run in a single batch",
)
@click.option(
    "--temperature",
    "-t",
    type=click.FLOAT,
    default=0.5,
    help="max_tokens passed to the model",
)
@click.option(
    "--engine",
    "-e",
    type=click.STRING,
    default="text-davinci-003",
    help="model as accepted by the openai API",
)
@click.option("--key", "-k", multiple=True, help="The keys to run")
@click.option(
    "--output",
    "-o",
    type=click.Path(),
)
@click.option("--silent", "-s", is_flag=True)
def run(
    path,
    verbose,
    force,
    dry_run,
    style,
    temperature,
    max_tokens,
    engine,
    key,
    output,
    silent,
    repair,
    human,
    shuffle,
    limit,
):
    """Run some prompts/suites!"""
    click.secho("💡 ¡promptimize! 💡", fg="cyan")
    if dry_run:
        click.secho("# DRY RUN MODE ACTIVATED!", fg="red")
    uses_cases = discover_objects(path, BasePromptCase)
    completion_create_kwargs = {
        "engine": engine,
        "max_tokens": max_tokens,
        "temperature": temperature,
    }

    report = None
    if output:
        report = Report.from_path(output)

    suite = Suite(uses_cases, completion_create_kwargs)
    suite.execute(
        verbose=verbose,
        style=style,
        silent=silent,
        report=report,
        dry_run=dry_run,
        keys=key,
        force=force,
        repair=repair,
        human=human,
        shuffle=shuffle,
        limit=limit,
    )

    if output:
        output_report = Report.from_suite(suite)
        if report:
            output_report.merge(report)
        click.secho(f"# Writing file output to {output}", fg="yellow")
        output_report.write(output, style=style)


cli.add_command(run)


@click.command(help="report on how your suites of prompts are performing")
@click.argument(
    "path",
    required=True,
    type=click.Path(exists=True),
)
@click.option("--groupby", "-g", help="GROUPBY", default="category")
def report(path, groupby):
    """Get some summary of how your prompt suites are performing"""
    click.secho(f"# Reading report @ {path}", fg="yellow")
    report = Report.from_path(path)
    report.print_summary(groupby)


cli.add_command(report)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from typing import List

project = "promptimize"
copyright = "2023, Maxime Beauchemin"
author = "Maxime Beauchemin"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions: List = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",  # Optional, for Google and NumPy-style docstrings
    "recommonmark",  # If you're using the recommonmark extension
    "sphinx_click.ext",
]


templates_path = ["_templates"]
exclude_patterns: List = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_logo = "https://user-images.githubusercontent.com/487433/229948453-36cbc2d1-e71f-4e87-9111-ab428bc96f4c.png"
html_static_path = ["_static"]



"""
This module provides a Suite class to manage and execute a collection of
use cases (prompts) to be tested. It allows running the tests, displaying
results, and serializing the summary of the suite.
"""
import random
from typing import Any, Dict, List, Optional, Union

import click

from promptimize import utils
from promptimize.prompt_cases import BasePromptCase


def separator(fg=None) -> None:
    """Print a separator line."""
    click.secho("# " + "-" * 40, fg=fg)


def separated_section(s, fg=None):
    separator(fg)
    click.secho(s, fg=fg)
    separator(fg)


class Suite:
    """A collection of use cases to be tested.

    Attributes:
        name (Optional[str]): The name of the suite.
        prompts (Dict[str, Prompt]): Dictionary of prompts to be tested,
            keyed by the prompt key.
        last_run_completion_create_kwargs (Dict[str, Any]): Keyword arguments
            used in the last run for completion creation.
    """

    def __init__(
        self,
        prompts: List["BasePromptCase"],
        name: Optional[str] = None,
    ) -> None:
        """
        Args:
            prompts (List[Prompt]): List of prompts to be tested.
            name (Optional[str]): The name of the suite. Defaults to None.
        """
        self.name = name
        self.prompts = {o.key: o for o in prompts}
        self.last_run_completion_create_kwargs: dict = {}

    def execute(  # noqa
        self,
        verbose: bool = False,
        style: str = "yaml",
        silent: bool = False,
        report=None,
        dry_run: bool = False,
        keys: list = None,
        force: bool = False,
        repair: bool = False,
        human: bool = False,
        shuffle: bool = False,
        limit: int = 0,
    ) -> None:
        """
        Execute the suite with the given settings.

        Args:
            verbose (bool): If True, print verbose output. Defaults to False.
            style (str): Output style for serialization. Defaults to "yaml".
            silent (bool): If True, suppress output. Defaults to False.
        """
        prompts = list(self.prompts.values())
        if keys:
            prompts = [p for p in prompts if p.key in keys]
        if repair and report:
            failed_keys = report.failed_keys
            prompts = [p for p in prompts if p.key in failed_keys]

        if shuffle:
            random.shuffle(prompts)

        if limit:
            prompts = prompts[:limit]

        for i, prompt in enumerate(prompts):
            should_run = force or self.should_prompt_execute(prompt, report)
            progress = f"({i+1}/{len(prompts)})"
            if not silent:
                if should_run:
                    separated_section(f"# {progress} [RUN] prompt: {prompt.key}", fg="cyan")
                else:
                    separated_section(f"# {progress} [SKIP] prompt: {prompt.key}", fg="yellow")

            if should_run:
                prompt._run(dry_run)
                if not dry_run:
                    prompt.test()

            if not silent and should_run:
                prompt.print(verbose=verbose, style=style)

            if should_run and human:
                v = click.prompt(
                    'Press Enter to continue, "Y" to force success, "N" to force fail, "X" to exit',
                    default="",
                    show_default=False,
                )
                v = v.lower()
                if v == "":
                    click.secho("Leaving result unaltered", fg="yellow")
                elif v == "y":
                    prompt.execution.score = 1
                    prompt.execution.human_override = True
                    click.secho("Forcing SUCCESS", fg="green")
                elif v == "n":
                    prompt.execution.score = 0
                    prompt.execution.human_override = True
                    click.secho("Forcing FAILURE", fg="red")
                elif v == "x":
                    break

        # `self.last_run_completion_create_kwargs = completion_create_kwargs
        if not silent:
            separated_section("# Suite summary", fg="cyan")
            click.echo(utils.serialize_object(self._serialize_run_summary(), style))

    def should_prompt_execute(self, prompt, report):
        if not report or not report.prompts:
            return True
        report_prompt = report.prompts.get(prompt.key)
        if not report_prompt:
            return True
        else:
            if not report_prompt.execution:
                return True

        if report_prompt.prompt_hash == prompt.prompt_hash:
            return False

        return True

    def _serialize_run_summary(
        self, verbose: bool = False
    ) -> Dict[str, Union[Optional[float], Dict[str, Any]]]:
        """
        Serialize the run summary of the suite.

        Args:
            verbose (bool): If True, include verbose output. Defaults to False.

        Returns:
            Dict[str, Union[Optional[float], Dict[str, Any]]]: Serialized run summary of the suite.
        """
        prompts = self.prompts.values()
        tested = [p for p in prompts if p.was_tested and p.execution.score is not None]
        suite_score = None
        if len(tested) > 0:
            total_weight = sum([p.weight for p in tested])
            suite_score = sum([p.execution.score for p in tested]) / total_weight
        d = {
            "suite_score": suite_score,
            "git_info": utils.get_git_info(),
        }

        return d

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the suite to a dictionary.

        Returns:
            Dict[str, Any]: Dictionary representation of the suite.
        """
        return {
            "name": self.name,
            "prompts": {p.key: p.to_dict() for p in self.prompts.values()},
            "run_summary": self._serialize_run_summary(),
        }



import sys
import importlib
import pkgutil
from pathlib import Path
from typing import List, Type, Any


def is_instance_or_derivative(obj: Any, object_type: Type) -> bool:
    return isinstance(obj, object_type)


def discover_objects(path: str, object_type: Type) -> List[Any]:  # noqa
    objects = []
    folder_path = Path(path).resolve()

    def process_module(module):
        # Iterate over the objects in the module
        for name, obj in module.__dict__.items():
            # Check if the object is an instance or derivative of the specified type
            if is_instance_or_derivative(obj, object_type):
                objects.append(obj)
            # Check if the object is a list or tuple containing instances or
            # derivatives of the specified type
            elif isinstance(obj, (list, tuple)):
                for item in obj:
                    if is_instance_or_derivative(item, object_type):
                        objects.append(item)

    # If the path points to a file, import the module and process it directly
    if folder_path.is_file() and folder_path.suffix == ".py":
        sys.path.insert(0, str(folder_path.parent))
        module_name = folder_path.stem
        module = importlib.import_module(module_name)
        process_module(module)

    # If the path points to a directory, proceed with the existing logic
    elif folder_path.is_dir():
        # Add the folder to the Python path to enable importing modules from it
        if folder_path not in sys.path:
            sys.path.insert(0, str(folder_path))

        # Iterate over all the modules in the folder
        for _, module_name, _ in pkgutil.iter_modules([str(folder_path)]):
            # Import the module
            module = importlib.import_module(module_name)
            process_module(module)

    return objects



from setuptools import setup, find_packages

# Read the contents of the requirements.txt file
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

# Read the contents of the README.md file
with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="promptimize",
    version="0.2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "promptimize=promptimize:cli",
            "p9e=promptimize:cli",
        ],
    },
    author="Maxime Beauchemin",
    author_email="maximebeauchemin@gmail.com",
    description="A python toolkit to generate and evaluate prompts for GPT at scale",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="Apache License, Version 2.0",
    license_file="LICENSE",
)



import yaml
from box import Box

import pandas as pd

from promptimize import utils


class Report:
    """Report objects interacting with the filesystem / databases and data structures"""

    version = "0.1.0"

    def __init__(self, path=None, data=None):
        self.data = Box()
        if data:
            self.data = Box(data)
        self.path = path

    def write(self, path=None, style="yaml"):
        """write the report to the filesystem"""
        path = path or self.path
        with open(path, "w") as f:
            f.write(utils.serialize_object(self.data.to_dict(), highlighted=False, style=style))

    def merge(self, report):
        """merge in another report into this one"""
        all_keys = set(report.prompts.keys()) | set(self.prompts.keys())
        for k in all_keys:
            a = report.prompts.get(k)
            b = self.prompts.get(k)
            if a and b:
                if a.execution.get("run_at", "") > b.execution.get("run_at", ""):
                    self.prompts[k] = a
                else:
                    self.prompts[k] = b

            if not a:
                self.prompts[k] = b
            elif not b:
                self.prompts[k] = a

    @property
    def prompts(self):
        """list the prompts in this report"""
        if self.data:
            return self.data.prompts
        return {}

    @property
    def failed_keys(self):
        """return the list of prompt keys that have not suceeded"""
        keys = set()
        for p in self.prompts.values():
            if p.execution.get("score", 0) < 1:
                keys.add(p.key)
        return keys

    @classmethod
    def from_path(cls, path):
        """load a report object from a path in the filesystem"""
        try:
            with open(path, "r") as f:
                report = cls(path, yaml.safe_load(f))
            report.fix_stuff()
            return report
        except FileNotFoundError:
            return None

    @classmethod
    def from_suite(cls, suite):
        """load a report object from a suite instance"""
        report = cls(data=suite.to_dict())
        return report

    def fix_stuff(self):
        # DELETEME
        for p in self.prompts.values():
            p.execution.score = 1 if p.execution.get("is_identical") else 0
            if p.execution.get("test_results_avg"):
                del p.execution.test_results_avg
        self.write()

    def get_prompt(self, prompt_key):
        """get a specific prompt data structure from the report"""
        return self.prompts.get(prompt_key)

    def prompt_df(self):
        """make a flat pandas dataframe out of the prompts in the reports"""
        prompts = [p for p in self.prompts.values() if p.execution]
        return pd.json_normalize(prompts)

    def print_summary(self, groupby="category"):
        """print the summary from the report"""
        if groupby:
            self.print_summary(groupby=None)

        df = self.prompt_df()

        df["score"] = df["weight"] * df["execution.score"]

        if groupby:
            df = df[[groupby, "weight", "score"]].groupby(groupby).sum()
        else:
            df = df.agg({"weight": "sum", "score": "sum"}).to_frame().T
        df["perc"] = (df["score"] / df["weight"]) * 100
        df = df.sort_values(by="weight", ascending=False)
        headers = []
        if groupby:
            headers = "keys"
        else:
            df = df.T
        print(utils.trabulate(df, headers=headers))
